@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 #   Mozo Menu Editor - Simple fd.o Compliant Menu Editor
 #   Copyright (C) 2006  Travis Watkins, Heinrich Wendel
+#   Copyright (C) 2012-2021 MATE Developers
 #
 #   This library is free software; you can redistribute it and/or
 #   modify it under the terms of the GNU Library General Public
@@ -18,7 +19,6 @@
 
 import codecs
 import os
-import re
 import xml.dom.minidom
 import xml.parsers.expat
 import locale
@@ -34,7 +34,10 @@ class Menu:
     dom = None
 
 class MenuEditor(object):
-    def __init__(self):
+
+    def __init__(self, undo_button, redo_button):
+        self.undo_button = undo_button
+        self.redo_button = redo_button
         self.locale = locale.getdefaultlocale()[0]
         self.__undo = []
         self.__redo = []
@@ -53,6 +56,7 @@ class MenuEditor(object):
         self.settings.tree.connect('changed', self.menuChanged)
         self.load()
         self.__loadMenus()
+        self.update_undo_redo_button_state()
 
     def __loadMenus(self):
         self.applications.path = os.path.join(util.getUserMenuPath(), self.applications.tree.props.menu_basename)
@@ -122,6 +126,7 @@ class MenuEditor(object):
             util.removeWhitespaceNodes(menu.dom)
         #reset undo/redo, no way to recover from this
         self.__undo, self.__redo = [], []
+        self.update_undo_redo_button_state()
         self.save()
 
     def revertTree(self, menu):
@@ -162,30 +167,39 @@ class MenuEditor(object):
             pass
         self.save()
 
+    def update_undo_redo_button_state(self):
+        self.redo_button.set_sensitive(len(self.__redo) > 0)
+        self.undo_button.set_sensitive(len(self.__undo) > 0)
+
     def undo(self):
         if len(self.__undo) == 0:
             return
         files = self.__undo.pop()
         redo = []
-        for file_path in files:
-            new_path = file_path.rsplit('.', 1)[0]
+        for undo_path in files[::-1]:
+            new_path = undo_path.rsplit('.', 1)[0]
+            if not os.path.exists(undo_path):
+                continue
             redo_path = util.getUniqueRedoFile(new_path)
 
-            f_file_path = codecs.open(new_path, 'r', 'utf-8')
-            f_new_path = codecs.open(new_path, 'rw', 'utf-8')
-            f_redo_path = codecs.open(redo_path, 'rw', 'utf-8')
+            # create redo file
+            try:
+                with codecs.open(new_path, 'r', 'utf-8') as f_new:
+                    with codecs.open(redo_path, 'w', 'utf-8') as f_redo:
+                        f_redo.write(f_new.read())
+                redo.append(redo_path)
+            except FileNotFoundError:
+                pass
 
-            data = f_new_path.read()
-            f_redo_path.write(data)
-            data = f_file_path.read()
-            f_new_path.write(data)
+            # restore undo file
+            try:
+                with codecs.open(undo_path, 'r', 'utf-8') as f_undo:
+                    with codecs.open(new_path, 'w', 'utf-8') as f_new:
+                        f_new.write(f_undo.read())
+                os.unlink(undo_path)
+            except FileNotFoundError:
+                pass
 
-            f_file_path.close()
-            f_new_path.close()
-            f_redo_path.close()
-
-            os.unlink(file_path)
-            redo.append(redo_path)
         # reload DOM to make changes stick
         for name in ('applications', 'settings'):
             menu = getattr(self, name)
@@ -194,28 +208,39 @@ class MenuEditor(object):
             except (IOError, xml.parsers.expat.ExpatError):
                 menu.dom = xml.dom.minidom.parseString(util.getUserMenuXml(menu.tree))
             util.removeWhitespaceNodes(menu.dom)
-        self.__redo.append(redo)
+        if redo:
+            self.__redo.append(redo)
+        self.update_undo_redo_button_state()
 
     def redo(self):
         if len(self.__redo) == 0:
             return
         files = self.__redo.pop()
         undo = []
-        for file_path in files:
-            new_path = file_path.rsplit('.', 1)[0]
+        for redo_path in files[::-1]:
+            new_path = redo_path.rsplit('.', 1)[0]
+            if not os.path.exists(redo_path):
+                continue
             undo_path = util.getUniqueUndoFile(new_path)
-            f_file_path = codecs.open(new_path, 'r', 'utf-8')
-            f_new_path = codecs.open(new_path, 'rw', 'utf-8')
-            f_undo_path = codecs.open(undo_path, 'rw', 'utf-8')
-            data = f_new_path.read()
-            f_undo_path.write(data)
-            data = f_file_path.read()
-            f_new_path.write(data)
-            os.unlink(file_path)
-            undo.append(undo_path)
-            f_file_path.close()
-            f_new_path.close()
-            f_undo_path.close()
+
+            # create undo file
+            try:
+                with codecs.open(new_path, 'r', 'utf-8') as f_new:
+                    with codecs.open(undo_path, 'w', 'utf-8') as f_undo:
+                        f_undo.write(f_new.read())
+                undo.append(undo_path)
+            except FileNotFoundError:
+                pass
+
+            # restore redo file
+            try:
+                with codecs.open(redo_path, 'r', 'utf-8') as f_redo:
+                    with codecs.open(new_path, 'w', 'utf-8') as f_new:
+                        f_new.write(f_redo.read())
+                os.unlink(redo_path)
+            except FileNotFoundError:
+                pass
+
         #reload DOM to make changes stick
         for name in ('applications', 'settings'):
             menu = getattr(self, name)
@@ -224,7 +249,9 @@ class MenuEditor(object):
             except (IOError, xml.parsers.expat.ExpatError):
                 menu.dom = xml.dom.minidom.parseString(util.getUserMenuXml(menu.tree))
             util.removeWhitespaceNodes(menu.dom)
-        self.__undo.append(undo)
+        if undo:
+            self.__undo.append(undo)
+        self.update_undo_redo_button_state()
 
     def getMenus(self, parent=None):
         if parent is None:
@@ -236,28 +263,35 @@ class MenuEditor(object):
             while item_type != MateMenu.TreeItemType.INVALID:
                 if item_type == MateMenu.TreeItemType.DIRECTORY:
                     item = item_iter.get_directory()
-                    yield (item, self.__isVisible(item))
+                    if item.get_menu_id() != 'Collection':
+                        yield (item, self.__isVisible(item))
                 item_type = item_iter.next()
 
     def getContents(self, item):
+        class CollectionDirectoryException(Exception): pass
+
         contents = []
         item_iter = item.iter()
         item_type = item_iter.next()
 
         while item_type != MateMenu.TreeItemType.INVALID:
             item = None
-            if item_type == MateMenu.TreeItemType.DIRECTORY:
-                item = item_iter.get_directory()
-            elif item_type == MateMenu.TreeItemType.ENTRY:
-                item = item_iter.get_entry()
-            elif item_type == MateMenu.TreeItemType.HEADER:
-                item = item_iter.get_header()
-            elif item_type == MateMenu.TreeItemType.ALIAS:
-                item = item_iter.get_alias()
-            elif item_type == MateMenu.TreeItemType.SEPARATOR:
-                item = item_iter.get_separator()
-            if item:
-                contents.append(item)
+            try:
+                if item_type == MateMenu.TreeItemType.DIRECTORY:
+                    item = item_iter.get_directory()
+                    if item.get_menu_id() == 'Collection': raise CollectionDirectoryException()
+                elif item_type == MateMenu.TreeItemType.ENTRY:
+                    item = item_iter.get_entry()
+                elif item_type == MateMenu.TreeItemType.HEADER:
+                    item = item_iter.get_header()
+                elif item_type == MateMenu.TreeItemType.ALIAS:
+                    item = item_iter.get_alias()
+                elif item_type == MateMenu.TreeItemType.SEPARATOR:
+                    item = item_iter.get_separator()
+                if item:
+                    contents.append(item)
+            except CollectionDirectoryException:
+                pass
             item_type = item_iter.next()
         return contents
 
@@ -376,14 +410,12 @@ class MenuEditor(object):
         file_path = item.get_desktop_file_path()
         keyfile = GLib.KeyFile()
         keyfile.load_from_file(file_path, util.KEY_FILE_FLAGS)
-
         util.fillKeyFile(keyfile, dict(Categories=[], Hidden=False))
 
-        app_info = item.get_app_info()
-        file_id = util.getUniqueFileId(app_info.get_name().replace(os.sep, '-'), '.desktop')
+        file_id = util.getUniqueFileId(os.path.basename(file_path).split(".desktop", 1)[0], ".desktop")
         out_path = os.path.join(util.getUserItemPath(), file_id)
 
-        contents, length = keyfile.to_data()
+        contents = keyfile.to_data()[0]
 
         with open(out_path, 'w') as f:
             f.write(contents)
@@ -397,15 +429,19 @@ class MenuEditor(object):
     def moveItem(self, item, new_parent, before=None, after=None):
         undo = []
         if item.get_parent() != new_parent:
-            #hide old item
+            # create new item
+            self.copyItem(item, new_parent)
+            # hide old item
             self.deleteItem(item)
-            undo.append(item)
-            file_id = self.copyItem(item, new_parent)
-            item = ('Item', file_id)
-            undo.append(item)
-        self.__positionItem(new_parent, item, before, after)
+        else:
+            self.__positionItem(new_parent, item, before, after)
         undo.append(self.__getMenu(new_parent))
         self.__addUndo(undo)
+        if item.get_parent() != new_parent:
+            # merge the undo entries created by the previous operations in the
+            # correct order so we can undo in one go
+            items = self.__undo.pop()[::-1] + self.__undo.pop()[::-1] + self.__undo.pop()[::-1]
+            self.__undo.append(items[::-1])
         self.save()
 
     def moveMenu(self, menu, new_parent, before=None, after=None):
@@ -503,6 +539,7 @@ class MenuEditor(object):
             with codecs.open(undo_path, 'w', 'utf-8') as f:
                 f.write(data)
             self.__undo[-1].append(undo_path)
+        self.update_undo_redo_button_state()
 
     def __getMenu(self, item):
         root = item.get_parent()
@@ -653,7 +690,7 @@ class MenuEditor(object):
         else:
             file_id = util.getUniqueFileId(keyfile.get_string(GLib.KEY_FILE_DESKTOP_GROUP, 'Name'), '.desktop')
 
-        contents, length = keyfile.to_data()
+        contents = keyfile.to_data()[0]
 
         with open(os.path.join(util.getUserItemPath(), file_id), 'w') as f:
             f.write(contents)
@@ -673,7 +710,7 @@ class MenuEditor(object):
 
         util.fillKeyFile(keyfile, kwargs)
 
-        contents, length = keyfile.to_data()
+        contents = keyfile.to_data()[0]
 
         with open(os.path.join(util.getUserDirectoryPath(), file_id), 'w') as f:
             f.write(contents)
@@ -753,10 +790,6 @@ class MenuEditor(object):
         xml_parent = self.__getXmlMenu(self.__getPath(parent), dom.documentElement, dom)
         self.__addXmlFilename(xml_parent, dom, file_id, 'Include')
 
-    def moveItem(self, parent, item, before=None, after=None):
-        self.__positionItem(parent, item, before=before, after=after)
-        self.save()
-
     def __positionItem(self, parent, item, before=None, after=None):
         contents = self.getContents(parent)
         if after:
